/*******************************************************************************
**
**  Module  : tpro_drv26.c
**  Date    : 03/31/06
**  Purpose : This driver provides an interface to the TPRO-PCI
**
**  Copyright(C) 2006 Spectracom Corporation. All Rights Reserved.
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
**
********************************************************************************
**
++     05/28/07 Modifications for multi-processor support
++     01/23/07 Added support for multiple simultaneous users
**     03/31/06 Updated from Linux (2.4.24) to Linux (2.6.11.12)
**
*******************************************************************************/
#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif
#include <linux/config.h>
#ifdef CONFIG_SMP
#define __SMP__
#endif

// #define TPRO_DEBUG 1

/*******************************************************************************
**          Includes
*******************************************************************************/
#include <linux/module.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <asm/uaccess.h>
#include <linux/interrupt.h>
#include <linux/cdev.h>

#include "ddtpro.h"
#include "tpropci.h"
#include "tpro_func.h"
#include "os_wait.h"

/*******************************************************************************
**          Defines
*******************************************************************************/
#define VERSION           "2.30"
#define IO_BASE_ADDR_REG  (3)
#define IO_BASE_ADDR_REG_32  (1)

/*******************************************************************************
**          Kernel Module Description
*******************************************************************************/
MODULE_AUTHOR("Spectracom Corporation (spectracomcorp.com)");
MODULE_DESCRIPTION("Spectracom TPRO/TSAT-PCI Timing Board");
MODULE_SUPPORTED_DEVICE("PCI Timing Board");
MODULE_LICENSE("GPL");

/*******************************************************************************
**          Brief Identification Strings
*******************************************************************************/
char tpropci_driver_name[]    = TPRO_PCI_DRV_NAME;
char tpropci_driver_string[]  = TPRO_PCI_DRV_STRING;
char tpropci_driver_version[] = VERSION;
char tpropci_copyright[]      = TPRO_PCI_DRV_COPYRIGHT;

/*******************************************************************************
**          Local Function Prototypes
*******************************************************************************/

void           tpropci_cleanup     (void);

int            tpropci_close       (struct inode *inode, struct file *filp);

tpropci_dev_t* tpropci_find_device (struct inode *inode);

int            tpropci_init        (void);

int            tpropci_init_one    (struct pci_dev             *pdev,
                                    const struct pci_device_id *ent);

int            tpropci_ioctl       (struct inode  *inode,
                                    struct file   *filp,
                                    unsigned int   cmd,
                                    unsigned int  arg);

int            tpropci_open        (struct inode *inode, struct file *filp);

void           tpropci_remove_one  (struct pci_dev *pdev);

static irqreturn_t tpropci_isr(int irq, void *dev_id, struct pt_regs *regs);

/*******************************************************************************
**          Typedef's and Structures
*******************************************************************************/

/*
** PCI Device table
*/
static struct pci_device_id tpropci_pci_tbl[] __initdata =
{
    {
        vendor:      TPRO_VENDOR_ID,
        device:      TPRO_5V_DEVICE_ID,
        subvendor:   TPRO_SUBSYSTEM_VENDOR_ID,
        subdevice:   PCI_ANY_ID,
        class:       0,
        class_mask:  0,
        driver_data: 0
    },
    {
        vendor:      TPRO_VENDOR_ID,
        device:      TPRO_3V_DEVICE_ID,
        subvendor:   TPRO_SUBSYSTEM_VENDOR_ID,
        subdevice:   PCI_ANY_ID,
        class:       0,
        class_mask:  0,
        driver_data: 0
    },
    {
        vendor:      TPRO_U66_VENDOR_ID,
        device:      TPRO_U66_DEVICE_ID,
        subvendor:   TPRO_U66_SUBSYSTEM_VENDOR_ID,
        subdevice:   PCI_ANY_ID,
        class:       0,
        class_mask:  0,
        driver_data: 0
    },
    {0, }
};

MODULE_DEVICE_TABLE(pci, tpropci_pci_tbl);

/*
**  tpropci_driver PCI driver information structure
*/
static struct pci_driver tpropci_driver =
{
    name:     TPRO_PCI_DRV_NAME,
    id_table: tpropci_pci_tbl,
    probe:    tpropci_init_one,
    remove:   __devexit_p(tpropci_remove_one),
};

/*
** Driver setup structure
*/
static struct file_operations tpropci_fops =
{
    owner:   THIS_MODULE,
    ioctl:   tpropci_ioctl,
    open:    tpropci_open,
    release: tpropci_close
};

/*
** Device Model
*/
static struct   class_simple  *tpropci_devClassp; /* Needed to use sysfs.     */
static struct   cdev           tpropci_cdev;     /* Character dev struct.     */
static          dev_t          tpropci_devNum;   /* Driver's base Dev Number  */
static          tpropci_dev_t *tpropci_devp[TPRO_PCI_MAX_NUM];
static          char           NameStr[TPRO_PCI_MAX_NUM][32];


// Define the 16 register interface functions
extern FuncTable_t tpropci_16bit;

// Define the 32 register interface functions
extern FuncTable_t tpropci_32bit;

/*******************************************************************************
**
** Function: tpropci_cleanup()
**
** Description: Routine called just before the driver is removed from memory.
**
** Parameters:
**     IN: None
**
**     RETURNS: None
**
*******************************************************************************/
void tpropci_cleanup(void)
{
    /* Unregister the driver with the PCI subsystem */
    pci_unregister_driver(&tpropci_driver);

    class_simple_destroy(tpropci_devClassp);
    cdev_del(&tpropci_cdev);
    unregister_chrdev_region(tpropci_devNum, TPRO_PCI_MAX_NUM);

    TPRO_PRINT("%s removed\n", tpropci_driver_string);

} /* End - tpropci_cleanup() */

/*******************************************************************************
**
** Function: tpropci_close()
**
** Description: Close a tpropci device
**
** Parameters:
**     IN: *inode - inode pointer
**         *filp  - file pointer
**
**     RETURNS: -EIO - Couldn't find matching tpropci device
**                 0 - Success
**
*******************************************************************************/
int tpropci_close (struct inode *inode, struct file *filp)
{
    tpropci_user_t * user = NULL;
    tpropci_dev_t *  hw = NULL;
    int status = 0;

    /* Retrieve the user structure from the file pointer. */
    user = (tpropci_user_t *)(filp->private_data);
    filp = NULL;
    if (user == NULL)
    {
        status = -EFAULT;
	TPRO_LOG_ERR;
        goto fxnExit;
    }

    /* Retrieve the device pointer from the user structure. */
    hw = user->pDevice;
    if (hw == NULL)
    {
        status = -EFAULT;
	TPRO_LOG(TPRO_LOG_ERROR,
                 ("[J%lu]%s: Null device pointer in user struct %p!!!\n",
		 jiffies, __FUNCTION__, user));
    }

    #ifdef TPRO_CHECK
        /* Retrieve the device pointer from the inode and verify. */
        if (hw != tpropci_find_device(inode))
        {
            status = -EFAULT;
            TPRO_LOG(TPRO_LOG_ERROR,
                     ("[J%lu]%s: Device pointer mismatch"
                      " in user struct %p!!!\n",
                      jiffies, __FUNCTION__, user));
            goto fxnExit;
        }
    #endif

    /*
    ** Remove the user structure from the list and deallocate it.
    ** If this is the only user, shut down the device.
    */

    TPRO_SPIN_LOCK(&hw->userLock);

    #ifdef TPRO_CHECK
        /* Make sure the first and last user pointers are not NULL. */
        if ((hw->pFirstUser == NULL) || (hw->pLastUser == NULL))
        {
            TPRO_SPIN_UNLOCK(&hw->userLock);
            status = -EFAULT;
            TPRO_LOG(TPRO_LOG_ERROR,
                     ("[J%lu]%s: hw->pFirstUser=%p, hw->pLastUser=%p in"
                      " device struct %p!!!\n",
                     jiffies, __FUNCTION__, hw->pFirstUser, hw->pLastUser,
                     user));
            goto fxnExit;
        }
    #endif

    if (user == hw->pFirstUser)
    {
        if (user == hw->pLastUser)
	{
	    /* This is the only user. */
            #ifdef TPRO_CHECK
                if ((user->pPrev != user) || (user->pNext != user))
		{
		    TPRO_LOG(TPRO_LOG_ERROR,
                             ("[J%lu]%s: Only user is %p but pPrev = %p"
                              " and pNext = %p",
                             jiffies, __FUNCTION__, user, user->pPrev,
                             user->pNext));
	        }
            #endif
            hw->pFirstUser = NULL;
            hw->pLastUser = NULL;
        }
        else
        {
            /* This is the first, but not the only, user. */
            user->pPrev->pNext = hw->pFirstUser = user->pNext;
            user->pNext->pPrev = user->pPrev;
        }

    }
    else if (user == hw->pLastUser)
    {
        /* This user was added last and is not the only user. */
        user->pNext->pPrev = hw->pLastUser = user->pPrev;
        user->pPrev->pNext = user->pNext;
    }
    else
    {
        /* Neither first nor last user.  Remove from the middle of the list. */
        user->pNext->pPrev = user->pPrev;
        user->pPrev->pNext = user->pNext;
    }

    TPRO_SPIN_UNLOCK(&hw->userLock);

    /* Free the user structure. */
    kfree(user);

    TPRO_LOG(TPRO_LOG_CLOSE,
             ("[J%lu]%s: Closed user struct %p\n",
              jiffies, __FUNCTION__, user));

 fxnExit:
    return status;

} /* End - tpropci_close() */

/*******************************************************************************
**
** Function: tpropci_find_device()
**
** Description: Determines which tpropci_dev_t stucture the inode is referring
**              to and return a pointer to it.  Returns NULL if no matching
**              tpropci_dev_t could be found.
**
** Parameters:
**     IN: *inode - info to match with a tpropci_dev_t sturcture
**
**     RETURNS: tpropci_dev_t* - NULL if none is found.
**
*******************************************************************************/
tpropci_dev_t* tpropci_find_device(struct inode *inode)
{
    tpropci_dev_t *retDevp = NULL;
    unsigned int   idx;

    for (idx = 0; idx < TPRO_PCI_MAX_NUM; idx++)
    {
        if (tpropci_devp[idx] != NULL)
        {
            if (tpropci_devp[idx]->devNum == inode->i_rdev)
            {
                retDevp = tpropci_devp[idx];
            }
        }
    }

    return retDevp;

} /* End - tpropci_find_device() */

/*******************************************************************************
**
** Function: tpropci_init()
**
** Description: TPRO kernel module entry point
**
** Parameters:
**     IN: None
**
**     RETURNS: 0 on success, or negative error code.
**
*******************************************************************************/
int tpropci_init(void)
{
    int rv;

    /* Print the driver ID strings. */
    TPRO_PRINT("%s - version %s\n%s\n",
               tpropci_driver_string,
               tpropci_driver_version,
               tpropci_copyright);

    rv = alloc_chrdev_region(&tpropci_devNum,
                             TPRO_PCI_BASE_MINOR,
                             TPRO_PCI_MAX_NUM,
                             TPRO_PCI_DRV_NAME);

    if (rv < 0)
    {
        TPRO_LOG(TPRO_LOG_ERROR,
                 (KERN_ALERT "[J%lu]%s: FAILED alloc_chrdev_region() err=%d\n",
		 jiffies, __FUNCTION__, rv));
        return rv;
    }

    /*
    ** Initialize char device structure
    */
    cdev_init(&tpropci_cdev, &tpropci_fops);
    tpropci_cdev.owner = THIS_MODULE;
    tpropci_cdev.ops   = &tpropci_fops;

    /*
    ** Add the char device to the system
    */
    rv = cdev_add(&tpropci_cdev, tpropci_devNum, TPRO_PCI_MAX_NUM);

    if (rv)
    {
        TPRO_LOG(TPRO_LOG_ERROR,
                 (KERN_ERR "[J%lu]%s: FAILED cdev_add()\n",
                 jiffies, __FUNCTION__));
        unregister_chrdev_region(tpropci_devNum, TPRO_PCI_MAX_NUM);
        return rv;
    }

    /*
    ** Create class and devices in sysfs
    */
    tpropci_devClassp = class_simple_create(THIS_MODULE, TPRO_PCI_DRV_NAME);

    if (IS_ERR(tpropci_devClassp))
    {
        TPRO_LOG(TPRO_LOG_ERROR,
                 (KERN_ERR "[J%lu]%s: Error creating sysfs class.\n",
		 jiffies, __FUNCTION__));
        rv = (int)PTR_ERR(tpropci_devClassp);
        cdev_del(&tpropci_cdev);
        unregister_chrdev_region(tpropci_devNum, TPRO_PCI_MAX_NUM);
        return rv;
    }

    /*
    **  register the driver with the PCI subsytem
    */
    rv = pci_register_driver(&tpropci_driver);

    return rv;

} /* End - tpropci_init() */

/*******************************************************************************
**
** Function: tpropci_init_one()
**
** Description: Initialize new device to be handled by this driver
**
** Parameters:
**     IN: *pdev - PCI device structure pointer
**         *ent  - PCI device ID pointer
**
**     RETURNS: 0 - Success, <0 - Failure
**
*******************************************************************************/
int tpropci_init_one(struct pci_dev *pdev, const struct pci_device_id *ent)
{
    int                  rv;
    struct class_device *class_err;
    unsigned int         idx;


    TPRO_PRINT("tpropci_init_one\n");

    /* Find the next available device pointer... */
    for (idx = 0; idx < TPRO_PCI_MAX_NUM; idx++)
    {
        /* If the currently indexed pointer hasn't been allocated yet... */
        if (tpropci_devp[idx] == NULL)
        {
            /* Have found available device pointer */
            break;
        }
    }

    /* If all device pointers have been already allocated... */
    if (idx >= TPRO_PCI_MAX_NUM)
    {
        TPRO_LOG_ERR;
        return -ENOSPC;
    }

    /* Allocate device structure for new device */
    tpropci_devp[idx] = kmalloc (sizeof(tpropci_dev_t), GFP_KERNEL);

    /* If memory couldn't be allocated... */
    if (tpropci_devp[idx] == NULL)
    {
        TPRO_LOG(TPRO_LOG_ERROR,
                 (KERN_ALERT "[J%lu]%s: No memory\n", jiffies, __FUNCTION__));
        return -ENOMEM;
    }

    TPRO_LOG(TPRO_LOG_LOAD,
             ("[J%lu]%s: Created device structure %p at index %d\n",
             jiffies, __FUNCTION__, tpropci_devp[idx], idx));

    /*
    **  prepare device driver access context
    */
    tpropci_devp[idx]->intCnt     = 0;
    tpropci_devp[idx]->pFirstUser = NULL;
    tpropci_devp[idx]->pLastUser  = NULL;
    tpropci_devp[idx]->headIndex = 0;

    TPRO_MUTEX_INIT(&tpropci_devp[idx]->deviceLock);
    TPRO_SPIN_LOCK_INIT(&tpropci_devp[idx]->userLock);

    /*
    **  set PCI configuration register to enable PCI I/O and Memory access
    */
    if (pci_enable_device(pdev))
    {
        if (tpropci_devp[idx]) kfree(tpropci_devp[idx]);
        TPRO_LOG(TPRO_LOG_ERROR,
                 (KERN_ALERT "[J%lu]%s: Enable PCI device failed\n",
                 jiffies, __FUNCTION__));
        return -EIO;
    }

    /*
    **  get base address of device and amount of I/O space requested
    */
    if (pdev->device == TPRO_U66_DEVICE_ID)
    {
        tpropci_devp[idx]->ioAddr  = pci_resource_start(pdev, IO_BASE_ADDR_REG_32);
        tpropci_devp[idx]->ioSize  = pci_resource_len(pdev, IO_BASE_ADDR_REG_32);
    }
    else
    {
        tpropci_devp[idx]->ioAddr  = pci_resource_start(pdev, IO_BASE_ADDR_REG);
        tpropci_devp[idx]->ioSize  = pci_resource_len(pdev, IO_BASE_ADDR_REG);
    }

    tpropci_devp[idx]->options = pdev->subsystem_device;

    /* Store irq in local object */
    tpropci_devp[idx]->irq     = pdev->irq;

    /* Set device number in the tpropci device structure */
    tpropci_devp[idx]->devNum  = MKDEV(MAJOR(tpropci_devNum), idx);

    sprintf(NameStr[idx], "tpropci%d", idx);

    /* Select the function set based on the board type. */
    /* The selection is based on the Device ID, it is one value */
    /* for the 32 bit version of the boards and one of two */
    /* for the 16 bit boards. */
    if (pdev->device == TPRO_U66_DEVICE_ID)
    {
        TPRO_PRINT("TPro_PCI 32 bit version of board\n");
        tpropci_devp[idx]->pFunctionTable = &tpropci_32bit;
    }
    else
    {
        TPRO_PRINT("TPro_PCI 16 bit version of board\n");
        tpropci_devp[idx]->pFunctionTable = &tpropci_16bit;
    }

    /* Add device into the tpropci class of devices */
    class_err = class_simple_device_add(tpropci_devClassp,
                                        MKDEV(MAJOR(tpropci_devNum), idx),
                                        NULL,
                                        NameStr[idx]);

    if (IS_ERR(class_err))
    {
        rv = PTR_ERR(class_err);
        class_simple_destroy(tpropci_devClassp);
        cdev_del(&tpropci_cdev);
        unregister_chrdev_region(tpropci_devNum, TPRO_PCI_MAX_NUM);
        TPRO_LOG(TPRO_LOG_ERROR,
                 (KERN_ALERT "[J%lu]%s: device add failed\n",
                 jiffies, __FUNCTION__));
        return rv;
    }

    sprintf(NameStr[idx], "tpropci%d", MINOR(tpropci_devp[idx]->devNum));

    /* Map device IRQ */
    rv = request_irq(tpropci_devp[idx]->irq,
                     tpropci_isr,
                     SA_INTERRUPT | SA_SHIRQ,
                     NameStr[idx],
                     tpropci_devp[idx]);
    if (rv != 0)
    {
        TPRO_LOG(TPRO_LOG_ERROR,
	         ("[J%lu]%s: Could NOT map IRQ!!!\n",
                 jiffies, __FUNCTION__));
        return -EIO;
    }

    /* Initialize the board.  Used to be done in IOCTL. */
    (void)tpropci_devp[idx]->pFunctionTable->InitializeBoard(tpropci_devp[idx]);

    return 0;

} /* End - tpropci_init_one() */

/*******************************************************************************
**
** Function: tpropci_ioctl()
**
** Description: Called when an ioctl operation is requested.
**
** Parameters:
**     IN: *inode - inode struct for the device
**         *filp  - file struct for the device
**          cmd   - the ioctl command
**          arg   - typecast for the arg, can be a value/ptr
**
**     RETURNS: Negative on failure or the ioctl call return value (if any)
**
*******************************************************************************/
int  tpropci_ioctl(struct inode  *inode,
                   struct file   *filp,
                   unsigned int   cmd,
                   unsigned long  arg)
{
    tpropci_ioctl_cmd_t *cmdHnd     = (tpropci_ioctl_cmd_t*)arg;
    tpropci_dev_t       *hw         = NULL;
    tpropci_user_t      *user       = NULL;
    int                  status     = 0;

    /* Validate the command. */
    if (_IOC_NR(cmd) >= IOCTL_TPRO_CMD_COUNT)
    {
        status = -EFAULT;
        TPRO_LOG(TPRO_LOG_ERROR,
		 ("[J%lu]%s: [CMD:0x%08X, USER:%p]: Invalid IOCTL command\n",
		 jiffies, __FUNCTION__, cmd, user));
        goto fxnExit;
    }

    /* Retrieve the user structure from the file pointer. */
    user = (tpropci_user_t *)(filp->private_data);
    if (user == NULL)
    {
        status = -EFAULT;
	TPRO_LOG_ERR;
        goto fxnExit;
    }

    TPRO_LOG(TPRO_LOG_IOCTL,
             ("[J%lu]%s: [CMD:0x%08X, USER:%p, HW:%p]\n",
             jiffies, __FUNCTION__, cmd, user, user->pDevice));

    /* Retrieve the device pointer from the user structure. */
    hw = user->pDevice;
    if (hw == NULL)
    {
        status = -EFAULT;
        TPRO_LOG(TPRO_LOG_ERROR,
                 ("[J%lu]%s: [CMD:0x%08X, USER:%p]: Null device pointer"
                  " in user struct!!!\n",
		  jiffies, __FUNCTION__, cmd, user));
        goto fxnExit;
    }

    #ifdef TPRO_CHECK
        /* Retrieve the device pointer from the inode and verify. */
        if (hw != tpropci_find_device(inode))
        {
 	    TPRO_LOG(TPRO_LOG_ERROR,
                     ("[J%lu]%s: [CMD:0x%08X, USER:%p, HW:%p]"
                      ": Device pointer mismatch!!!\n",
                     jiffies, __FUNCTION__, cmd, user, user->pDevice));
        }
    #endif

    /* Verify access to the command object (read/write) */
    if (!access_ok(VERIFY_WRITE, (void*)cmdHnd, sizeof(tpropci_ioctl_cmd_t)))
    {
        status = -EFAULT;
        goto fxnExit;
    }

    /* Make sure the call is currently available to the user. */
    if ((cmd == IOCTL_TPRO_GET_TIME) || (cmd == IOCTL_TPRO_OPEN) ||
        (cmd == IOCTL_TPRO_SYNCH_STATUS) || (cmd == IOCTL_TPRO_GET_DRIVER) ||
        (cmd == IOCTL_TPRO_PEEK) || (cmd == IOCTL_TPRO_POKE))
    {
        /* These commands are available to everyone. */
    }
    else
    {
        int firstUser;

        /* These commands are available only to the first user. */
        TPRO_SPIN_LOCK(&hw->userLock);
        firstUser = (hw->pFirstUser == user) ? 1 : 0;
        TPRO_SPIN_UNLOCK(&hw->userLock);
	if (firstUser == 0)
        {
            status = -EFAULT;
            TPRO_LOG(TPRO_LOG_INFO_1,
                     ("[J%lu]%s: [CMD:0x%08X, USER:%p, HW:%p]"
                      ": Multiuser conflict!!!\n",
		     jiffies, __FUNCTION__, cmd, user, user->pDevice));
            goto fxnExit;
	}
    }

    switch (cmd)
    {
    case IOCTL_TPRO_OPEN:
    {
        BoardObj board;

        /* verify access to board object (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof(BoardObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* copy object from user space */
        if (copy_from_user(&board, &cmdHnd->argVec[0], sizeof (BoardObj))) {
            status = -EFAULT;
            goto fxnExit;
        }

        /* copy board options to board object */
        board.options = hw->options;

        /* copy local object to user object */
        if (copy_to_user(&cmdHnd->argVec[0], &board, sizeof (BoardObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_GET_ALTITUDE:
    {
        AltObj drvAlt;

        /* verify access to altitude object (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof(AltObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* get altitude from tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->getAltitude(hw, &drvAlt);
	TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* copy object to user space */
        if (copy_to_user(&cmdHnd->argVec[0], &drvAlt, sizeof (AltObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_GET_DATE:
    {
        DateObj drvDate;

        /* verify access to date object (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof(DateObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* get date from tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->getDate(hw, &drvDate);
	TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* copy object to user space */
        if (copy_to_user(&cmdHnd->argVec[0], &drvDate, sizeof(DateObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_GET_DRIVER:
    {
        char version[7];

        /* verify access to version string (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof(version)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* copy version to local array */
        strcpy(version, VERSION);

        /* copy string to user space */
        if (copy_to_user(&cmdHnd->argVec[0], version, sizeof (version)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_GET_FIRMWARE:
    {
        unsigned char firmware[5];

        /* verify access to firmware bytes (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof(firmware)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* get firmware from the tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->getFirmware(hw, firmware);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* copy string to user space */
        if (copy_to_user(&cmdHnd->argVec[0], firmware, sizeof (firmware)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_GET_LATITUDE:
    {
        LatObj drvLat;

        /* verify access to latitude object (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof(LatObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* get latitude from tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->getLatitude(hw, &drvLat);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* copy object to user space */
        if (copy_to_user(&cmdHnd->argVec[0], &drvLat, sizeof (LatObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_GET_LONGITUDE:
    {
        LongObj drvLong;

        /* verify access to longitude object (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof (LongObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* get longitude from tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->getLongitude(hw, &drvLong);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* copy object to user space */
        if (copy_to_user(&cmdHnd->argVec[0], &drvLong, sizeof (LongObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_GET_SAT_INFO:
    {
        SatObj drvSat;

        /* verify access to satellite info object (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof (SatObj))) {
            status = -EFAULT;
            goto fxnExit;
        }

        /* get satellite info from tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->getSatInfo(hw, &drvSat);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* copy object to user space */
        if (copy_to_user(&cmdHnd->argVec[0], &drvSat, sizeof (SatObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_GET_TIME:
    {
        TimeObj drvTime;

        /* verify access to time object (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof (TimeObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* get time from tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->getTime(hw, &drvTime);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* copy object to user space */
        if (copy_to_user(&cmdHnd->argVec[0], &drvTime, sizeof (TimeObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_RESET_FIRMWARE:
    {
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->resetFirmware(hw);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);
        break;
    }

    case IOCTL_TPRO_SET_HEARTBEAT:
    {
        HeartObj drvHeart;

        /* copy object from user space */
        if (copy_from_user(&drvHeart, &cmdHnd->argVec [0], sizeof (HeartObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* set heartbeat on tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->setHeartbeat(hw, &drvHeart);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);
        break;
    }

    case IOCTL_TPRO_SET_MATCHTIME:
    {
        MatchObj drvMatch;

        /* copy object from user space */
        if (copy_from_user(&drvMatch, &cmdHnd->argVec[0], sizeof (MatchObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* set match time on tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->setMatchTime(hw, &drvMatch);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);
        break;
    }

    case IOCTL_TPRO_SET_PROP_DELAY_CORR:
    {
        int microseconds;

        /* If parameter is NULL, just return success. */
        if (cmdHnd == NULL)
	{
	    break;
	}

        /* get microsecond value from user */
        __get_user(microseconds, (int *) &cmdHnd->argVec[0]);

        /* set propagation delay correction value on tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->setPropDelayCorr(hw, &microseconds);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);
        break;
    }

    case IOCTL_TPRO_SET_TIME:
    {
        TimeObj drvTime;

        /* copy object from user space */
        if (copy_from_user(&drvTime, &cmdHnd->argVec [0], sizeof (TimeObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* set time on tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->setTime(hw, &drvTime);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);
        break;
    }

    case IOCTL_TPRO_SET_YEAR:
    {
        unsigned short year;

        /* get year value from user */
        __get_user(year, (unsigned short *) &cmdHnd->argVec[0]);

        /* set year to tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->setYear(hw, &year);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);
        break;
    }

    case IOCTL_TPRO_SIM_EVENT:
    {
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->simEvent(hw);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);
        break;
    }

    case IOCTL_TPRO_SYNCH_CONTROL:
    {
        int ctrl;

        /* get control value from user */
        __get_user(ctrl, (int *) &cmdHnd->argVec[0]);

        /* set synchronization control value to tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->synchControl(hw, (unsigned char*)&ctrl);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);
        break;
    }

    case IOCTL_TPRO_SYNCH_STATUS:
    {
        unsigned char stat = 0xFF;

        /* verify access to synch status character (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof (unsigned char)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* get synchronization status from tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->synchStatus(hw, &stat);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* copy status to user-space */
        if (copy_to_user(&cmdHnd->argVec[0], &stat, sizeof (unsigned char)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_WAIT_EVENT:
    {
        WaitObj drvWaitEvent;

        /* verify access to wait object (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof (WaitObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* copy object from user space */
        if (copy_from_user(&drvWaitEvent,
                           &cmdHnd->argVec[0],
                           sizeof (WaitObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->toggleInterrupt(hw, FIFO_IRQ_ENABLE);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* wait for event if necessary */
        if (hw->headIndex == hw->tailIndex)
        {
            /* clear flag condition */
            atomic_set (&hw->eventFlag, 0);

            /* wait on semaphore */
            status = os_waitPend(&hw->event_wait,
                                 drvWaitEvent.jiffies,
                                 &hw->eventFlag);

            /* return on timeout or waitaphore error */
            if (status == OS_WAIT_STATUS_TIMEOUT)
            {
                TPRO_MUTEX_LOCK(&hw->deviceLock);
                (void)hw->pFunctionTable->toggleInterrupt(hw, DISABLE_INTERRUPTS);
                TPRO_MUTEX_UNLOCK(&hw->deviceLock);
                return 1;
            }

            if (status == OS_WAIT_STATUS_NG)
            {
                TPRO_MUTEX_LOCK(&hw->deviceLock);
                (void)hw->pFunctionTable->toggleInterrupt(hw, DISABLE_INTERRUPTS);
                TPRO_MUTEX_UNLOCK(&hw->deviceLock);
                return -1;
            }
        }

        /* read event buffer into object */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->toggleInterrupt(hw, DISABLE_INTERRUPTS);
        (void)hw->pFunctionTable->getEvent(hw, &drvWaitEvent);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* copy object back to user space */
        if (copy_to_user(&cmdHnd->argVec[0], &drvWaitEvent, sizeof (WaitObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_WAIT_HEART:
    {
        int ticks;

        /* get microsecond value from user */
        __get_user(ticks, (int *) &cmdHnd->argVec[0]);

        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->toggleInterrupt(hw, HEART_IRQ_ENABLE);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* clear flag condition */
        atomic_set(&hw->heartFlag, 0);

        /* wait on semaphore */
        status = os_waitPend(&hw->heart_wait, ticks, &hw->heartFlag);

        if (status == OS_WAIT_STATUS_TIMEOUT) status = 1;
        if (status == OS_WAIT_STATUS_NG) status = -1;

        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->toggleInterrupt(hw, DISABLE_INTERRUPTS);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);
        break;
    }

    case IOCTL_TPRO_WAIT_MATCH:
    {
        int ticks;

        /* get microsecond value from user */
        __get_user(ticks, (int *) &cmdHnd->argVec[0]);

        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->toggleInterrupt(hw, MATCH_IRQ_ENABLE);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* clear flag condition */
        atomic_set(&hw->matchFlag, 0);

        /* wait on semaphore */
        status = os_waitPend(&hw->match_wait, ticks, &hw->matchFlag);

        if (status == OS_WAIT_STATUS_TIMEOUT) status = 1;
        if (status == OS_WAIT_STATUS_NG) status = -1;

        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->toggleInterrupt(hw, DISABLE_INTERRUPTS);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);
        break;
    }

    case IOCTL_TPRO_PEEK:
    {
        MemObj Reg;

        /* verify access to memory object (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof(MemObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* copy object from user space */
        if (copy_from_user(&Reg, &cmdHnd->argVec [0], sizeof (MemObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* get call the peek routine */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->peek(hw, &Reg);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* copy object back to user space */
        if (copy_to_user(&cmdHnd->argVec[0], &Reg, sizeof (MemObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_POKE:
    {
        MemObj Reg;

        /* copy object from user space */
        if (copy_from_user(&Reg, &cmdHnd->argVec [0], sizeof (MemObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* call the poke routine */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->poke(hw, &Reg);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);
        break;
    }

    case IOCTL_TPRO_GET_NTP_TIME:
    {
        NtpTimeObj ntpTime;

        /* verify access to NTP time object (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof (NtpTimeObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* get time from tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->getNtpTime(hw, &ntpTime);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* If subsystem device id is a TPRO family reference */
        if (hw->options < TSAT_PCI)
        {
            memcpy(&ntpTime.refId, "IRIG", 4);
        }
        else /* Else, TSAT reference */
        {
            memcpy(&ntpTime.refId, "GPS ", 4);
        }

        /* copy object to user space */
        if (copy_to_user(&cmdHnd->argVec[0], &ntpTime, sizeof(NtpTimeObj)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    case IOCTL_TPRO_GET_FPGA:
    {
        unsigned char fpgaVer[6];

        /* verify access to firmware bytes (write) */
        if (!access_ok(VERIFY_WRITE,
                       (void*)&cmdHnd->argVec[0],
                       sizeof(fpgaVer)))
        {
            status = -EFAULT;
            goto fxnExit;
        }

        /* get firmware from the tpro card */
        TPRO_MUTEX_LOCK(&hw->deviceLock);
        (void)hw->pFunctionTable->getFpgaVersion(hw, fpgaVer);
        TPRO_MUTEX_UNLOCK(&hw->deviceLock);

        /* copy string to user space */
        if (copy_to_user(&cmdHnd->argVec[0], fpgaVer, sizeof (fpgaVer)))
        {
            status = -EFAULT;
            goto fxnExit;
        }
        break;
    }

    default:
	TPRO_LOG(TPRO_LOG_INFO_1,
             ("[J%lu]%s: [CMD:0x%08X, USER:%p, HW:%p]"
              ": Unhandled command\n",
             jiffies, __FUNCTION__, cmd, user, user->pDevice));
        status = 1;
        goto fxnExit;
    }

    fxnExit:
        return status;

} /* End - tpropci_ioctl() */

/*******************************************************************************
**
** Function:    tpropci_isr()
** Description: The tpropci driver ISR routine
**
** Parameters:
**     IN:  irq    - interrupt
**         *dev_id - tpropci device pointer
**         *regs   - processor's context
**
**     RETURNS: IRQ_HANDLED - IRQ was handled properly
**
*******************************************************************************/
static irqreturn_t tpropci_isr(int irq, void *dev_id, struct pt_regs *regs)
{
    tpropci_dev_t* hw = (tpropci_dev_t *) dev_id;
    unsigned short status;
    MemObj Reg;

    /* read the interrupt status register */
    Reg.offset = hw->pFunctionTable->cmdStatOffset;
    hw->pFunctionTable->peek(hw, &Reg);
    status = Reg.value;

    /* check if the interrupt is ours */
    /* match int mask bit 6 */
    /* match Flag bit 3 */
    /* heartbeat int mask bit 5 */
    /* heartbeat Flag bit 4 */
    /* FIFO ready int mask bit 7 */
    /* FIFO empty Flag bit 0 */
    if (((status & 0x30) != 0x30) && /* heartbeat int and heartbeat flag */
        ((status & 0x48) != 0x48) && /* match int and match flag*/
        ((status & 0x81) != 0x80))   /* fifo Int and fifo rdy */
    {
        return IRQ_NONE;
    }

    /* If the heartbeat interrupt... */
    if (status & 0x10)
    {
        Reg.offset = hw->pFunctionTable->clrFlagsOffset;
        Reg.value = 0x10;
        Reg.l_value = 0x10;
        hw->pFunctionTable->poke(hw, &Reg);

        /* if we are waiting for interrupt post notification */
        if (atomic_read(&hw->heartFlag) == 0)
        {
            atomic_inc(&hw->heartFlag);
            os_waitPost(&hw->heart_wait);
        }
    }

    /* If the match interrupt... */
    if (status & 0x8)
    {
        Reg.offset = hw->pFunctionTable->clrFlagsOffset;
        Reg.value = 0x08;
        Reg.l_value = 0x08;
        hw->pFunctionTable->poke(hw, &Reg);

        /* if we are waiting for interrupt post notification */
        if (atomic_read(&hw->matchFlag) == 0)
        {
            atomic_inc(&hw->matchFlag);
            os_waitPost(&hw->match_wait);
        }
    }

    /* If the time-tag interrupt... */
    if (!(status & 0x1))
    {
        unsigned short *ptr = &hw->fifoVector [hw->headIndex];
        unsigned char eventIndex;

        hw->eventCnt++;

        /* read event */
        for (eventIndex = 0; eventIndex < 10; eventIndex++)
        {
            /* wait for data to be ready */
            Reg.offset = hw->pFunctionTable->cmdStatOffset;
            do
            {
                hw->pFunctionTable->peek(hw, &Reg);
                status = Reg.value;
            }
            while (status & 0x01);

            /* read word */
            Reg.offset = hw->pFunctionTable->fifoIntCtrlOffset;
            hw->pFunctionTable->peek(hw, &Reg);
            ptr[eventIndex] = Reg.value;
        }

        /* increment head index */
        hw->headIndex = ((hw->headIndex + 10) % (sizeof(hw->fifoVector)/sizeof(hw->fifoVector[0])));

        /* if we are waiting for interrupt post notification */
        if (atomic_read(&hw->eventFlag) == 0)
        {
            atomic_inc(&hw->eventFlag);
            os_waitPost(&hw->event_wait);
        }
    }

    hw->intCnt++;

    return IRQ_HANDLED;

} /* End - tpropci_isr() */

/*******************************************************************************
**
** Function: tpropci_open()
**
** Description: Open the TPRO-PCI device.
**
** Parameters:
**     IN: *inode - inode struct for the device
**         *file  - file pointer to return TPRO-PCI object reference with
**
**     RETURNS: 0 on success, or negative error code.
**
*******************************************************************************/
int tpropci_open(struct inode *inode, struct file *filp)
{
    tpropci_dev_t * hw;
    tpropci_user_t *user;

    hw = tpropci_find_device(inode);

    if (hw == NULL)
    {
        TPRO_LOG(TPRO_LOG_ERROR,
		 ("[J%lu]%s: Couldn't find matching device!!!\n",
                 jiffies, __FUNCTION__));
        return -EIO;
    }

    /* Allocate user structure and store device pointer. */
    user = kmalloc (sizeof(tpropci_user_t), GFP_KERNEL);
    if (user == NULL)
    {
        TPRO_LOG(TPRO_LOG_ERROR,
		 (KERN_ALERT "[J%lu]%s: No memory\n",
                 jiffies, __FUNCTION__));
        return -ENOMEM;
    }
    user->pDevice = hw;

    /* Lock the user list access.  If this is the first user, the lock must be maintained
     * around the device initialization as well. */
    TPRO_SPIN_LOCK(&hw->userLock);

    if (hw->pFirstUser != NULL)
    {
        user->pNext = hw->pFirstUser;
        user->pPrev = hw->pLastUser;
        user->pNext->pPrev = user->pPrev->pNext = hw->pLastUser = user;
    }
    else
    {
        hw->intCnt = 0;
        hw->eventCnt = 0;
        hw->headIndex = hw->tailIndex = 0;
        user->pNext = user->pPrev = hw->pFirstUser = hw->pLastUser = user;
    }

    /* Release access to the user list. */
    TPRO_SPIN_UNLOCK(&hw->userLock);

    TPRO_LOG(TPRO_LOG_OPEN,
	     ("[J%lu]%s: Adding user struct %p to device %p\n",
             jiffies, __FUNCTION__, user, hw));

    /* set private data */
    filp->private_data = user;

    return 0;

} /* End - tpropci_open */

/*******************************************************************************
**
** Function: tpropci_remove_one()
**
** Description: Removes a TPRO PCI device from the system
**
** Parameters:
**     IN: *pdev - tpropci device pointer
**
**     RETURNS: None
**
*******************************************************************************/
void tpropci_remove_one (struct pci_dev *pdev)
{
    tpropci_dev_t      *hw = NULL;
    unsigned int  idx;

    for (idx = 0; idx < TPRO_PCI_MAX_NUM; idx++)
    {
        if (tpropci_devp[idx] != NULL)
        {
            if (tpropci_devp[idx]->irq == pdev->irq)
            {
                hw = tpropci_devp[idx];
                tpropci_devp[idx] = NULL;
                break;
            }
        }
    }

    if (hw)
    {
        /* Disable interrupts on the device. */
        (void)hw->pFunctionTable->toggleInterrupt(hw, DISABLE_INTERRUPTS);

        /* Free the IRQ. */
        free_irq(hw->irq, hw);

        class_simple_device_remove(hw->devNum);

        /* Destroy the spinlock and mutex. */
        TPRO_SPIN_LOCK_DESTROY(&hw->userLock);
        TPRO_MUTEX_DESTROY(&hw->deviceLock);

        TPRO_LOG(TPRO_LOG_UNLOAD,
                 ("[J%lu]%s: Removed device structure %p from index %d\n",
                 jiffies, __FUNCTION__, hw, idx));

        /* Free the device object. */
        kfree (hw);
    }

} /* End - tpropci_remove_one */

/*******************************************************************************
            Explicit Module Routine Declarations
*******************************************************************************/
module_init(tpropci_init);    /*-- init module ---------*/
module_exit(tpropci_cleanup); /*-- cleanup module ------*/
