/*******************************************************************************
**
**  Module: pps.c
**  Date:   12/02/2003
**  Purpose
**  Description: Code file pps.c
**
**                 Copyright 2002, Spectracom Corporation.
**                          ALL RIGHTS RESERVED
**
** Description:
**     The driver that runs the GPS 1-PPS and the Conditioned 1-PPS.  The PPS
**     API described in RFC 2783 is implemented through ioctls.  In this
**     implementation, the only time format supported is PPS_TMFMT_TSPEC
**
********************************************************************************
**
**     6/28/05 Ported from uClinux (2.4.20) to standard Linux (2.4.26) for the
**             HanSolo project.  [MG]
**
**     10/06/09 Ported from HanSolo project into TSync-PCI driver.  [DHS]
**
*******************************************************************************/

#ifdef TSYNC_NTP

/* Includes */
#include <linux/errno.h>    /* Definitions for uniform error code return.     */
#include <linux/fs.h>       /* Needed for all device drivers.                 */
#include <linux/interrupt.h>/* Needed for request_irq and free_irq            */
#include <linux/ioctl.h>    /* Needed for ioctl operations.                   */
#include <linux/ioport.h>   /* Needed to accesss I/O port regions.            */
#include <linux/kernel.h>   /* We're operating in kernel space.               */
#include <linux/module.h>   /* Needed for all Loadable Kernel Module.         */
#include <linux/sched.h>    /* Needed to request IRQs.                        */
#include <linux/types.h>    /* Needed for specifying data widths.             */
#include <linux/wait.h>     /* Needed to use waitqueue                        */

/* PPS-API INCLUDES */
#include <linux/time.h>
#include <linux/timex.h>
#include <linux/timepps.h>

#include <asm/io.h>         /* Needed for I/O port read/write (e.g. inw, outw)*/
#include <asm/system.h>     /* Needed for memory barriers. */
#include <asm/uaccess.h>    /* Functions that move data to/from user space. */

#include "tsyncpci.h"
#include "tsync_hw_func.h"

#include "tsync_pps.h"

/*******************************************************************************
** Defines
*******************************************************************************/
/* #define PPS_DEBUG */
#undef PPS_DEBUG

/*******************************************************************************
** Types
*******************************************************************************/

/*******************************************************************************
** Data Declarations
*******************************************************************************/
static struct pps    s_pps[TSYNC_PCI_MAX_NUM];

/*******************************************************************************
** Local Function Prototypes
*******************************************************************************/
static int          pps_create(struct file*, pps_handle_t*);
static int          pps_destroy(struct file*, pps_handle_t*);
static int          pps_setparams(struct file*, struct pps_params*);
static int          pps_getparams(struct file*, struct pps_params*);
static int          pps_getcap(struct file*, int *);
static int          pps_fetch(struct file*, struct pps_fetch_args*);
#ifdef CONFIG_NTP_PPS
static int          pps_kcbind(struct file*, struct pps_bind_args*);
#endif
/* R3: DISABLED, func. body defined in /usr/src/linux/drivers/char/ppsapi
** static int       pps_allowed_mode(const struct pps*, mode_t, int*);
*/

/*******************************************************************************
 * inline Private functions -- must be declared before it is used.
 ******************************************************************************/
/*******************************************************************************
**
** Function:    do_clock_gettime
** Description: Emulate the do_clocl_gettime in linux 2.4 PPS-kit with
**              do_gettimeofday.
**
**
** Parameters:
**     IN:      int                 - Clock type
**     IN/OUT:  struct timespec*    - Where to place the timestamp
**     OUT:     None
**     Returns: Supposed to return the offset since last update of "xtime"
**
*******************************************************************************/
static inline unsigned long do_clock_gettime(int dummy, struct timespec *ts)
{
	do_gettimeofday((struct timeval *) ts);
	ts->tv_nsec *= 1000;	/* microseconds to nanoseconds */
	return 0;
}

/*******************************************************************************
**
** Function:    hardpps
** Description: Emulate the hardpps function
**
**
** Parameters:
**     IN:      long   nsec         - The offset since last update of "xtime"
**     IN/OUT:  struct timespec*    - Where to place the timestamp
**     OUT:     None
**     Returns: Supposed to return the offset since last update of "xtime"
**
*******************************************************************************/
static inline void hardpps(const struct timespec *p_ts, long nsec)
{
    /* Dummy function. */
}

/*******************************************************************************
 * Public functions
 ******************************************************************************/

/*******************************************************************************
**
** Function:    pps_init
** Description: Initialize the 1pps devices.
**
** Parameters:
**     IN:      None
**     IN/OUT:  None
**     OUT:     None
**     Returns: None
**
*******************************************************************************/
int __init pps_init(void)
{
    int i;

    #ifdef CONFIG_NTP_PPS
        /* Set the kernel consumer hardpps to empty values */
        pps_kc_hardpps_dev = NULL;
        pps_kc_hardpps_mode = 0;
    #endif

    /* Initialize pps structs and register each pps device with OS */
    for (i = 0; i < TSYNC_PCI_MAX_NUM; i++)
    {
        s_pps[i].magic                  = PPSCLOCK_MAGIC;
        s_pps[i].state.parm.api_version = PPS_API_VERS_1;
        s_pps[i].state.parm.mode        = PPS_TSFMT_TSPEC;
        s_pps[i].state.cap              = PPS_TSFMT_TSPEC  | PPS_CANWAIT       |
                                          PPS_CAPTURECLEAR | PPS_OFFSETASSERT  |
                                          PPS_OFFSETCLEAR  | PPS_CAPTUREASSERT |
                                          PPS_ECHOASSERT   | PPS_ECHOCLEAR;

        init_waitqueue_head(&s_pps[i].state.ewait);
    }

    return(0);

} /* End pps_init */

/*******************************************************************************
**
** Function:    pps_init_one
** Description: Initialize the specified 1pps device specific HW.
**
** Parameters:
**     IN:      tsyncpci_dev_t *hw  - HW device structure
**     IN/OUT:  None
**     OUT:     None
**     Returns: None
**
*******************************************************************************/
int pps_init_one(tsyncpci_dev_t *hw)
{
    /* Unmask the PPS interrupt */
    return(set_hw_int_mask_by_index(hw, 0, 0));

} /* End pps_init_one */

/*******************************************************************************
**
** Function:    pps_exit
** Description: The exit function that will be called when this module is
**              unloaded.
**
** Parameters:
**     In:      None
**     In/Out:  None
**     Out:     None
**     Returns: None
**
** Notes:
**    The __exit macro causes the omission of the function when the module is
**    built into the kernel, and like __init, has no effect for
**    loadable modules. Built-in drivers don't need a cleanup function,
**    while loadable modules do.
**
*******************************************************************************/
int __exit pps_exit(void)
{
    return (0);
} /* End pps_exit */

/*******************************************************************************
**
** Function:    pps_open
** Description: The open function that will be called when this device is
**              opened.
**
** Parameters:
**     In:      inode *inode      - the inode for this device
**              file *filp        - the file pointer for this device
**     In/Out:  None
**     Out:     None
**     Returns: None
**
** Notes:
**
*******************************************************************************/
int pps_open(struct inode* inode, struct file* filp)
{
    int retVal = 0;
    struct pps *s_pps_p;

    /* === filp->private_data ASSIGNMENT === */
    /* The driver open() method defaults the filp->private_data to NULL,
    ** unless it is registerered using devfs_register().
    ** If it is not NULL to begin with, do not overwrite it.
    ** Otherwise, assign it to the correct private_data based on the device
    ** minor number. */
    s_pps_p = ((tsyncpci_user_t*)filp->private_data)->pPpsData;

    if (s_pps_p == NULL)
    {
        s_pps_p = &s_pps[iminor(inode)];

        /* Assign the private_data to the correct struct pps address */
        ((tsyncpci_user_t*)filp->private_data)->pPpsData = s_pps_p;
    }

    return (retVal);

} /* End pps_open */

/*******************************************************************************
**
** Function:    pps_release
** Description: The release function that will be called when this device is
**              closed / released.
**
** Parameters:
**     In:      inode *inode      - the inode for this device
**              file *filp        - the file pointer for this device
**     In/Out:  None
**     Out:     None
**     Returns: None
**
** Notes:
**
*******************************************************************************/
int pps_release(struct inode* inode, struct file* filp)
{
    /* Stub */
    return (0);

} /* End pps_release */

/*******************************************************************************
**
** Function:    pps_ioctl
** Description: The API described in RFC 2783 is implemented here through ioctl
**              system calls.  The ioctl command numbers are defined in
**              /usr/src/linux/include/linux/timepps.h.
**
** Parameters:
**     In:      inode *inode      - the inode for this device
**              file *filp        - the file pointer for this device
**              unsigned int cmd  - the ioctl command to perform
**     In/Out:  unsigned long arg - command specific argument
**     Out:     None
**     Returns: Negative on error, otherwise >= 0
**
*******************************************************************************/
int pps_ioctl(struct inode* inode, struct file* filp, unsigned int cmd,
              unsigned long arg)
{
    int retVal = 0;

    switch (cmd)
    {
        case PPS_IOC_CREATE:
            retVal = pps_create(filp, (pps_handle_t *)arg);
            break;
        case PPS_IOC_DESTROY:
            retVal = pps_destroy(filp, (pps_handle_t *)arg);
            break;
        case PPS_IOC_SETPARMS:
            retVal = pps_setparams(filp, (struct pps_params *)arg);
            break;
        case PPS_IOC_GETPARMS:
            retVal = pps_getparams(filp, (struct pps_params *)arg);
            break;
        case PPS_IOC_GETCAP:
            retVal = pps_getcap(filp, (int *)arg);
            break;
        case PPS_IOC_FETCH:
            retVal = pps_fetch(filp, (struct pps_fetch_args *)arg);
            break;
        case PPS_IOC_KC_BIND:
    #ifdef CONFIG_NTP_PPS
            retVal = pps_kcbind(filp, (struct pps_bind_args *)arg);
    #else
            retVal = -EOPNOTSUPP;
    #endif
            break;
        default: /* Invalid command given */
            retVal = -ENOTTY;
            break;
    }

    return (retVal);

} /* End pps_ioctl */


/*******************************************************************************
**
** Function:    pps_isr
** Description: Common interrupt handler for the pps sources.  When an
**              interrupt occurs, the timestamp is updated for this pps's state
**              and any processes waiting for this pps are woken up.
**
** Parameters:
**     In:      None
**     In/Out:  None
**     Out:     None
**     Returns: None
**
*******************************************************************************/
void pps_isr(int src)
{
    struct pps      *pp = &s_pps[src];
    struct timespec  ts;
    int              mode;

    /* Get timestamp */
    pp->state.ecount = do_clock_gettime(CLOCK_REALTIME, &pp->state.etime);

    mode                        = pp->state.parm.mode;
    ts                          = pp->state.etime;     /* Save the timestamp */
    pp->state.info.current_mode = mode;

    #ifdef CONFIG_NTP_PPS
        /* Only call hardpps if this is the proper device. */
        /* NEVER USE NTP DEBUG MODE! The printk calls ruin the pps accuracy. */
        if (pps_kc_hardpps_dev == (&s_pps[src]))
        {
            hardpps(&ts, pp->state.ecount);
        }
    #endif /* CONFIG_NTP_PPS */

    /* Record the assert events if in proper mode */
    if (mode & PPS_CAPTUREASSERT)
    {
        /* Add assert offset to the time stamp as needed */
        if (mode & PPS_OFFSETASSERT)
        {
            ts.tv_nsec += pp->state.parm.assert_offset.tv_nsec;

            /* Adjust values if tv_nsec < 0 or >= NSEC_PER_SEC */
            if (ts.tv_nsec >= NSEC_PER_SEC)
            {
                ts.tv_nsec -= NSEC_PER_SEC;
                ts.tv_sec++;
            }
            else if (ts.tv_nsec < 0)
            {
                ts.tv_nsec += NSEC_PER_SEC;
                ts.tv_sec--;
            }

            /* Reset ts to original timestamp */
            ts = pp->state.etime;
        }

        /* Save time stamp and update sequence */
        pp->state.info.assert_timestamp = ts;
        pp->state.info.assert_sequence++;
    }
    else if (mode & PPS_CAPTURECLEAR)
    {
        /* Record the clear events if in proper mode */
        /* Add clear offset to the time stamp as needed */
        if (mode & PPS_OFFSETCLEAR)
        {
            ts.tv_nsec += pp->state.parm.clear_offset.tv_nsec;

            /* Adjust values if tv_nsec < 0 or >= NSEC_PER_SEC */
            if (ts.tv_nsec >= NSEC_PER_SEC)
            {
                ts.tv_nsec -= NSEC_PER_SEC;
                ts.tv_sec++;
            }
            else if (ts.tv_nsec < 0)
            {
                ts.tv_nsec += NSEC_PER_SEC;
                ts.tv_sec--;
            }
        }

        /* Save time stamp and update sequence */
        pp->state.info.clear_timestamp = ts;
        pp->state.info.clear_sequence++;
    }

    /* Wake up processes waiting on this pps */
    if (waitqueue_active(&pp->state.ewait))
    {
        wake_up_interruptible(&pp->state.ewait);
    }

} /* End pps_isr */


/*******************************************************************************
 * Private Functions
 ******************************************************************************/

/*******************************************************************************
**
** Function:    pps_create
** Description: This is a stub function that just does sanity checks on the
**              arguments passed in.  There is little need for this function,
**              as the struct pps that is used for each pps device is static
**              and each pps signal is represented as a different file.
**
** Parameters:
**     In:      file* filp           - the file pointer for this device
**              pps_handle_t* handle - the handle for this pps device
**     In/Out:  None
**     Out:     None
**     Returns: Negative on error, 0 on success
**
*******************************************************************************/
static int pps_create(struct file* filp, pps_handle_t* handle)
{
    struct pps *pp = ((tsyncpci_user_t*)filp->private_data)->pPpsData;
    int         retVal = 0;

    /* Make sure proper device is being used */
    if (pp->magic != PPSCLOCK_MAGIC)
    {
        retVal = -EOPNOTSUPP;
    }

    return (retVal);

} /* End pps_create */

/*******************************************************************************
**
** Function:    pps_destroy
** Description: This is a stub function that just does sanity checks on the
**              arguments passed in.  There is little need for this function,
**              as the struct pps that is used for each pps device is static
**              and each pps signal is represented as a different file.
**
** Parameters:
**     In:      file* filp           - the file for this device
**              pps_handle_t* handle - the handle for this pps device
**     In/Out:  None
**     Out:     None
**     Returns: Negative on error, 0 on success
**
*******************************************************************************/
static int pps_destroy(struct file* filp, pps_handle_t* handle)
{
    struct pps *pp = ((tsyncpci_user_t*)filp->private_data)->pPpsData;
    int         retVal = 0;

    /* Make sure nobody is waiting on this pps */
    if (waitqueue_active(&pp->state.ewait))
    {
        retVal = -EBUSY;
    }

    return (retVal);

} /* End pps_destroy */

/*******************************************************************************
**
** Function:    pps_fetch
** Description: Retrieve the pps_info_t that contains the time stamps for the
**              next pulse.  The calling process is put to sleep until a pps
**              signal is asserted.
**
** Parameters:
**     In:      file* filp            - the file pointer for this device
**     In/Out:  None
**     Out:     pps_fetch_args* fargs - the arguments needed for a fetch
**     Returns: Negative on error, 0 on success
**
*******************************************************************************/
static int pps_fetch(struct file* filp, struct pps_fetch_args* fargs)
{
    struct pps            *pp = ((tsyncpci_user_t*)filp->private_data)->pPpsData;
    unsigned long          flags;
    int                    ret = 0;
    long                   timeout;
    struct pps_fetch_args  fetchArgs;

    /* Setup waitqueue. */
    DEFINE_WAIT(wait);

    if (copy_from_user(&fetchArgs, fargs, sizeof(struct pps_fetch_args)))
    {
        return (-EFAULT);
    }

    /* Disable all interrupts */
    local_irq_save(flags);

    /* Validate arguments */
    if ((pp == NULL) || (pp->magic != PPSCLOCK_MAGIC))
    {
        /* Restore all interrupts */
        local_irq_restore(flags);
        return (-EOPNOTSUPP);
    }

    /* If the correct timestamp format is not specified */
    if (fetchArgs.tsformat != PPS_TSFMT_TSPEC)
    {
        /* Restore all interrupts */
        local_irq_restore(flags);
        return (-EINVAL);
    }

    timeout = HZ * fetchArgs.timeout.tv_sec;

    timeout += fetchArgs.timeout.tv_nsec / (1000000000 / HZ);

    /* If a non-zero timeout has been specified */
    if (timeout != 0)
    {
        /* Add wait queue entry to the queue and set the process state */
        prepare_to_wait(&(pp->state.ewait), &wait, TASK_INTERRUPTIBLE);

        /* If no timeout was specified */
        if (fetchArgs.timeout.tv_sec == -1)
        {
            /* Wait forever */
            schedule();

            /* Flags are restored by the above sleep call */
            local_irq_save(flags);
        }
        else /* A non-zero timeout was specified */
        {
            /* Wait with specified timeout */
            timeout = schedule_timeout(timeout);

            /* Flags are restored by the above sleep call */
            local_irq_save(flags);

            /* If the above wait call timed out */
            if (timeout == 0)
            {
                ret = (-ETIMEDOUT);
            }
        }

        finish_wait(&(pp->state.ewait), &wait);

        if (signal_pending(current))
        {
            ret = (-EINTR);
        }

        /* If an error was detected */
        if (ret != 0)
        {
            /* Restore all interrupts */
            local_irq_restore(flags);
            return ret;
        }
    } /* End - if non-zero timeout was specified */

    fetchArgs.pps_info_buf = pp->state.info;

    /* Restore all interrupts */
    local_irq_restore(flags);

    ret = copy_to_user(fargs, &fetchArgs, sizeof(struct pps_fetch_args)) ?
          -EFAULT : 0;

    return ret;

} /* End pps_fetch */

/*******************************************************************************
**
** Function:    pps_setparams
** Description: Set the parameters to those specified in the pps_params passed
**              in.
**
** Parameters:
**     In:      file* filp          - the file pointer for this device
**              pps_params* params  - the params to set this driver to
**     In/Out:  None
**     Out:     None
**     Returns: Negative on success, 0 on failure
**
*******************************************************************************/
static int pps_setparams(struct file* filp, struct pps_params* params)
{
    struct pps         *pp = ((tsyncpci_user_t*)filp->private_data)->pPpsData;
    struct pps_params   kparams;
    unsigned long       flags;
    int                 must_bits;
    int                 may_bits;
    int                 retVal = 0;

    local_irq_save(flags);

    /* Validate arguments */
    if ((pp == NULL) || (params == NULL))
    {
        local_irq_restore(flags);
        return (-EINVAL);
    }

    /* Make sure proper device is being used */
    if (pp->magic != PPSCLOCK_MAGIC)
    {
        local_irq_restore(flags);
        return (-EOPNOTSUPP);
    }

    /* Copy the user space data to the local kernel space data */
    retVal = copy_from_user(&(kparams), params, sizeof(struct pps_params));
    /* If copy cannot get all the data, abort */
    if (retVal != 0)
    {
        local_irq_restore(flags);
        return (-EFAULT);
    }

    /* Make sure allowed bits are being set */
    if ((kparams.mode & ~pp->state.cap) != 0)
    {
        local_irq_restore(flags);
        return (-EINVAL);
    }

    /* Need to have a format code in params */
    if ((kparams.mode & (PPS_TSFMT_TSPEC | PPS_TSFMT_NTPFP)) == 0)
    {
        local_irq_restore(flags);
        kparams.mode |= PPS_TSFMT_TSPEC;

        /* Copy kparams back to params to update this change. If kparams is
        ** is changed in other places in this function, do the copy_to_user call
        ** after the last change operation */
        retVal = copy_to_user(params, &(kparams), sizeof(struct pps_params));
        if (retVal != 0)
        {
        local_irq_restore(flags);
        return (-EFAULT);
        }
    }

    may_bits = pps_allowed_mode(pp, filp->f_mode, &must_bits);

    /* Check if mode passed in is allowed */
    if (((kparams.mode & must_bits) != must_bits) ||
        ((kparams.mode & ~may_bits) != 0))
    {
        local_irq_restore(flags);
        return (-EPERM);
    }

    if (capable(CAP_SYS_TIME))
    {
        /* Allow setting offsets */
        pp->state.parm = kparams;
    }
    else
    {
        pps_params_t  *ppspp = &pp->state.parm;
        ppspp->api_version   = kparams.api_version;
        ppspp->mode          = kparams.mode;
    }

    local_irq_restore(flags);
    return (retVal);

} /* End pps_setparams */

/*******************************************************************************
**
** Function:    pps_getparams
** Description: Get the current parameters of the pps device.
**
** Parameters:
**     In:      file* filp          - the file for this device
**     In/Out:  None
**     Out:     pps_params* params  - the location to store the parameters
**     Returns: Negative on failure, 0 on success
**
*******************************************************************************/
static int pps_getparams(struct file* filp, struct pps_params *params )
{
    struct pps    *pp = ((tsyncpci_user_t*)filp->private_data)->pPpsData;
    unsigned long  flags;
    int            retVal = 0;

    local_irq_save(flags);

    /* Validate arguments */
    if ((pp == NULL) || (params == NULL))
    {
        local_irq_restore(flags);
        return (-EINVAL);
    }

    /* Make sure proper device is being used */
    if (pp->magic != PPSCLOCK_MAGIC)
    {
        local_irq_restore(flags);
        return (-EOPNOTSUPP);
    }

    /* Copy the data to user space */
    retVal = copy_to_user(params, &(pp->state.parm), sizeof(struct pps_params));
    if (retVal != 0)
    {
        local_irq_restore(flags);
        return (-EFAULT);
    }
    local_irq_restore(flags);
    return (retVal);

} /* End pps_getparams */

/*******************************************************************************
**
** Function:    pps_getcap
** Description: Return the capabilities of this pps device.  The capabilites
**              are a series of bits whose values are defined in
**              /usr/src/linux/include/linux/timepps.h.
**
** Parameters:
**     In:      file* filp  - the file for this device
**     In/Out:  None
**     Out:     int* caps   - the location to store the capabilities bits
**     Returns: Negative on error, 0 on success
**
*******************************************************************************/
static int pps_getcap(struct file* filp, int* caps)
{
    struct pps    *pp = ((tsyncpci_user_t*)filp->private_data)->pPpsData;
    unsigned long  flags;
    int            must_bits;           /* pps_allowed_mode requires 3rd arg */
    int            retVal = 0;
    int            kcaps  = 0;

    local_irq_save(flags);

    /* Validate arguments */
    if ((pp == NULL) || (caps == NULL))
    {
        local_irq_restore(flags);
        return (-EINVAL);
    }

    /* Make sure proper device is being used */
    if (pp->magic != PPSCLOCK_MAGIC)
    {
        local_irq_restore(flags);
        return (-EOPNOTSUPP);
    }

    /* Retrieve bits that may be set */
    kcaps = pps_allowed_mode(pp, filp->f_mode, &must_bits);

    /* Copy the data to user space */
    retVal = copy_to_user(caps, &kcaps, sizeof(int));
    if (retVal != 0)
    {
        local_irq_restore(flags);
        return (-EFAULT);
    }

    local_irq_restore(flags);
    return (retVal);

} /* End pps_getcap */

#ifdef CONFIG_NTP_PPS
/*******************************************************************************
**
** Function:    pps_kcbind
** Description: This function specifies whether this is the pps source for the
**              kernel consumer.  The variables pps_kc_hardpps_dev and
**              pps_kc_hardpps_mode are included from hardpps.h.
**
** Parameters:
**     In:      file* filp           - the file pointer for this device
**              pps_bind_args* bargs - specific args for a bind
**     In/Out:  None
**     Out:     None
**     Returns: Negative on error, otherwise 0 on success
**
*******************************************************************************/
static int pps_kcbind(struct file* filp, struct pps_bind_args* bargs)
{
    struct pps           *pp = ((tsyncpci_user_t*)filp->private_data)->pPpsData;
    struct pps_bind_args  kbargs;
    unsigned long         flags;
    int                   retVal = 0;

    local_irq_save(flags);

    /* Validate arguments */
    if ((pp == NULL) || (bargs == NULL))
    {
        local_irq_restore(flags);
        return (-EINVAL);
    }

    /* Make sure proper device is being used */
    if (pp->magic != PPSCLOCK_MAGIC)
    {
        local_irq_restore(flags);
        return (-EOPNOTSUPP);
    }

    /* Check file permissions */
    if ((filp->f_mode & FMODE_WRITE) == 0)
    {
        local_irq_restore(flags);
        return (-EBADF);
    }

    /* Copy the user space data to the local kernel space data */
    retVal = copy_from_user(&kbargs, bargs, sizeof(struct pps_bind_args));
    /* If copy cannot get all the data, abort */
    if (retVal != 0)
    {
        local_irq_restore(flags);
        return (-EFAULT);
    }

    /* Validate members of kbargs */
    if ((kbargs.tsformat != PPS_TSFMT_TSPEC)       ||
        ((kbargs.edge    & ~PPS_CAPTUREBOTH) != 0) ||
        (kbargs.consumer != PPS_KC_HARDPPS))
    {
        local_irq_restore(flags);
        return (-EINVAL);
    }

    /* Permission check */
    if (!capable(CAP_SYS_TIME))
    {
        local_irq_restore(flags);
        return (-EPERM);
    }

    /* If setting the edge to zero, we're removing the hardpps device. */
    if (kbargs.edge == 0)
    {
        /* Make sure either we are the hardpps device or nobody is */
        if ((pps_kc_hardpps_dev == NULL) ||
            (pps_kc_hardpps_dev == filp->private_data))
        {
            pps_kc_hardpps_mode = kbargs.edge;
            pps_kc_hardpps_dev  = NULL;
        }
        else
        {
            retVal = -EINVAL;
        }
    }
    else
    {
        /* Make sure either we are the hardpps device or nobody is. */
        if ((pps_kc_hardpps_dev == NULL) ||
            (pps_kc_hardpps_dev == filp->private_data))
        {
            pps_kc_hardpps_mode = kbargs.edge;
            pps_kc_hardpps_dev  = filp->private_data;
        }
        else
        {
            retVal = -EINVAL;
        }
    }

    local_irq_restore(flags);
    return (retVal);

} /* End pps_kcbind */
#endif /* CONFIG_NTP_PPS */

/*******************************************************************************
**
** Function:    pps_allowed_mode
** Description: Return allowed mode bits for given pps struct, file's mode,
**              and user.  Bits set in '*obligatory' must be set.  Returned
**              bits may be set.
**
** Parameters:
**     In:      struct pps* pps - the pps structure for this device
**              mode_t fmode    - the mode the file was opened using
**     In/Out:  None
**     Out:     int* obligatory - the bits that MUST be set
**     Returns: The an int containing the bits that may be set for the params
**              or an error code.
*******************************************************************************/
/* R3: DISABLED, CODE IS NOW DEFINED IN /usr/src/linux/drivers/char/ppsapi
static int pps_allowed_mode(const struct pps *pps, mode_t fmode,
                            int* obligatory )
{
    int cap = pps->state.cap;

    // Always read only
    cap &= ~PPS_CANWAIT;

    // Only supported format
    *obligatory = PPS_TSFMT_TSPEC;

    if ((fmode & FMODE_WRITE) == 0)
    {
        // If file is not writeable
        cap = *obligatory = pps->state.parm.mode;

    }
    else if (!capable(CAP_SYS_TIME))
    {
        // if may not manipulate system time
        int fixed_bits;

        fixed_bits   = PPS_OFFSETASSERT | PPS_OFFSETCLEAR;
        fixed_bits  &= pps->state.parm.mode;
        *obligatory |= fixed_bits;
    }
    return (cap);

} *//* End pps_allowed_mode */

/*******************************************************************************
 * Private Functions
 ******************************************************************************/

#endif /* TSYNC_NTP */
